"""Routines and classes for Answer Set Programming source code parsing.

"""

import arpeggio as ap

from clyngor.asp_parsing import asp_grammar


def parse_asp_program(asp_source_code:str, do=None) -> tuple:
    parse_tree = ap.ParserPython(asp_grammar()).parse(asp_source_code)
    return ap.visit_parse_tree(parse_tree, visitor=do or CodeAsTuple())


class CodeAsTuple(ap.PTNodeVisitor):
    def __init__(self, int_builder:type=int):
        super().__init__()
        self._int_builder = int_builder

    def visit_number(self, node, children):
        return self._int_builder(node.value)

    def visit_args(self, node, children):
        return tuple(children)

    def visit_text(self, node, children):
        text = tuple(children)[0]
        return 'text', str(text)

    def visit_ident(self, node, children):
        return str(node.value)

    def visit_term(self, node, children):
        if children[0] == 'not':
            type = '¬term'
            predicate, *args = children[1:]
        else:
            type = 'term'
            predicate, *args = children
        assert len(args) in {0, 1}
        args = args[0] if len(args) == 1 else args
        return type, predicate, tuple(args)

    def visit_forall(self, node, children):
        atom, *conditions = children
        assert len(atom) == 3
        assert atom[0] in {'term', '¬term'}
        type = '¬forall' if atom[0] == '¬term' else 'forall'
        return type, atom[1], atom[2], tuple(conditions)

    def visit_expression(self, node, children):
        return tuple(children)

    def visit_body(self, node, children):
        predicate, *args = children
        for child in children:
            assert len(child) == 1
        return tuple(child[0] for child in children)

    def visit_constraint(self, node, children):
        if len(children) == 1:
            children = children[0]
        else:
            raise ValueError("Constraint body has not 1 value, but {}: {}".format(len(children), children))
        return 'constraint', tuple(children)


    def visit_selection_nobody(self, node, children):
        downum, upnum, body = 0, None, ()
        if len(children) == 3:
            downum, expr, upnum = children
        elif len(children) == 1:
            expr = children[0]
        else:
            assert len(children) == 2
            if isinstance(children[0], tuple):  # second number given
                expr, upnum = children
            else:  # first number given
                upnum, expr = children
        return 'selection', downum, upnum, expr, body


    def visit_selection_body(self, node, children):
        downum, upnum, body = 0, None, ()
        if len(children) == 4:
            downum, expr, upnum, body = children
        elif len(children) == 2:  # no numbers
            expr, body = children
        else:  # one number is given ; which one ?
            assert len(children) == 3
            tuples = tuple(isinstance(child, tuple) for child in children)
            assert any(tuples)
            if tuples[0]:  # the second number is given
                assert not tuples[1]
                expr, upnum, body = children
            elif tuples[1]:  # the first number is given
                downum, expr, body = children
            else:  # no number given WTF
                assert False, "algorithm is false."
            # determine if body is given
        return 'selection', downum, upnum, expr, body

    def visit_head(self, node, children):
        predicate, *args = children
        return children

    def visit_rule(self, node, children):
        head, *body = children
        if len(head) == 1:  # general case: only one production
            head = head[0][1:]  # remove the 'term'
            type = 'rule'
        else:
            type = 'multirule'
        if len(body) == 1:
            body = body[0]
        else:
            raise ValueError("Body has not 1 value, but {}: {}".format(len(body), body))
        return (type, *tuple(head), tuple(body))

    def visit_program(self, node, children):
        for rule in children:
            yield rule




from timeit import timeit
from pprint import pprint
from functools import partial

from clyngor import ASP
from clyngor import asp_parsing


ASP_CODE = """
rel(a,c). rel(a,d). rel(b,d). rel(b,e).
obj(X):- rel(X,_) ; rel(X,Y): att(Y).
att(Y):- rel(_,Y) ; rel(X,Y): obj(X).
:- not obj(X):obj(X).
:- not att(Y):att(Y).
"""

def time_efficiency():
    func = partial(asp_parsing.program_to_dependancy_graph, ASP_CODE)
    print('Perform 1000 graphs in', timeit(func, number=1000), 'seconds.')
    # last time: 3.4s


if __name__ == '__main__':
    answers = ASP(ASP_CODE)
    for answer in answers.by_predicate.first_arg_only:
        print('{' + ','.join(answer['obj']) + '} × {' + ','.join(answer['att']) + '}')
    print()

    print('Dependancy graph:')
    pprint(asp_parsing.program_to_dependancy_graph(ASP_CODE))
    print()

    print('Benchmark:')
    time_efficiency()



"""Definition of the ASP grammar"""


import arpeggio as ap


def asp_grammar():
    """Implement the Answer Set Programming grammar.

    """

    def ident():      return ap.RegExMatch(r'[a-z][a-zA-Z0-9_]*')
    def number():     return ap.RegExMatch(r'-?[0-9]+')
    def text():       return '"', ap.RegExMatch(r'((\\")|([^"]))*'), '"'
    def litteral():   return [ident, text, number]
    def variable():   return ap.RegExMatch(r'(([A-Z][a-zA-Z0-9_]*)|(_))')
    def subterm():    return [(ident, ap.Optional("(", args, ")")), litteral, variable]
    def args():       return subterm, ap.ZeroOrMore(',', subterm)
    def term():       return ap.Optional('not'), ident, ap.Optional("(", args, ")")
    def forall():     return term, ':', term, ap.ZeroOrMore(',', term)
    def expression(): return [forall, term]
    def body():       return expression, ap.ZeroOrMore(';', expression)
    def constraint(): return ':-', body
    def head():       return term, ap.ZeroOrMore(';', term)
    def rule():       return head, ':-', body
    def selection_nobody():  return ap.Optional(number), '{', ap.OneOrMore(expression), '}', ap.Optional(number)
    def selection_body():  return ap.Optional(number), '{', ap.OneOrMore(expression), '}', ap.Optional(number), ':-', body
    def selection():  return [selection_body, selection_nobody]
    def instruction():return [selection, constraint, rule, term]
    def program():    return ap.OneOrMore(instruction, '.')

    return program





from clyngor.asp_parsing import parse_asp_program, CodeAsTuple


def test_constraint():
    ASP_CODE = r"""
    :- a.
    :- not obj(X):obj(X).
    """
    rules = tuple(parse_asp_program(ASP_CODE, do=CodeAsTuple()))
    assert len(rules) == 2
    first, second = rules

    print('FIRST:', first)
    expected = ('constraint', (('term', 'a', ()),) )
    print('EXPEC:', expected)
    assert first == expected

    print('SECOD:', second)
    expected = ('constraint', (('¬forall', 'obj', ('X',), (('term', 'obj', ('X',)),) ),) )
    print('EXPEC:', expected)
    assert second == expected


def test_forall():
    ASP_CODE = r"""
    a:- rel(X,Y): obj(X), att(Y).
    b:- not rel(X,Y): obj(X), att(Y).
    """
    rules = tuple(parse_asp_program(ASP_CODE, do=CodeAsTuple()))
    assert len(rules) == 2
    first, second = rules

    print('FIRST:', first)
    expected = ('rule', 'a', (), (('forall', 'rel', ('X', 'Y'), (('term', 'obj', ('X',)), ('term', 'att', ('Y',))) ),) )
    print('EXPEC:', expected)
    assert first == expected

    print('SECOD:', second)
    expected = ('rule', 'b', (), (('¬forall', 'rel', ('X', 'Y'), (('term', 'obj', ('X',)), ('term', 'att', ('Y',))) ),) )
    print('EXPEC:', expected)
    assert second == expected


def test_term():
    ASP_CODE = r"""
    a.
    b(1).
    c(2):- not b(2).
    """
    rules = tuple(parse_asp_program(ASP_CODE, do=CodeAsTuple()))
    assert len(rules) == 3
    first, second, third = rules

    print('FIRST:', first)
    expected = ('term', 'a', ())
    print('EXPEC:', expected)
    assert first == expected

    print('SECOD:', second)
    expected = ('term', 'b', (1,))
    print('EXPEC:', expected)
    assert second == expected

    print('THIRD:', third)
    expected = ('rule', 'c', (2,), (('¬term', 'b', (2,)),) )
    print('EXPEC:', expected)
    assert third == expected


def test_selection():
    ASP_CODE = r"""
    1 { sel(X): obj(X) } 1.
    { con(X,Y): obj(X), att(Y, 2) }.
    2 { sel(X): obj(X) } 4:- anatom.
    """
    rules = tuple(parse_asp_program(ASP_CODE, do=CodeAsTuple()))
    assert len(rules) == 3

    first, second, third = rules

    print('FIRST:', first)
    expected = ('selection', 1, 1, (('forall', 'sel', ('X',), (('term', 'obj', ('X',)),)),), ())
    print('EXPEC:', expected)
    assert first == expected

    print('SECOD:', second)
    expected = ('selection', 0, None, (('forall', 'con', ('X', 'Y'),
                                        (('term', 'obj', ('X',)),
                                         ('term', 'att', ('Y', 2)),),
                                        ),), ())
    print('EXPEC:', expected)
    assert second == expected

    print('THIRD:', third)
    expected = ('selection', 2, 4, (('forall', 'sel', ('X',),
                                     (('term', 'obj', ('X',)),),
                                     ),), (('term', 'anatom', ()),))
    print('EXPEC:', expected)
    assert third == expected


def test_multiple_rules():
    ASP_CODE = r"""
    a.
    b("les amis, \"coucou\".").
    obj(X):- rel(X,_) ; rel(X,Y): att(Y).
    att(Y):- rel(_,Y) ; rel(X,Y): obj(X).
    """

    rules = tuple(parse_asp_program(ASP_CODE))
    assert len(rules) == 4

    first, second, third, fourth = rules
    assert first == ('term', 'a', ())

    print('SECOND:', second)
    assert second == ('term', 'b', (('text', r'les amis, \"coucou\".'),))

    print('THIRD:', third)
    expected = ('rule', 'obj', ('X',), (
        ('term', 'rel', ('X', '_')),
        ('forall', 'rel', ('X', 'Y'), (('term', 'att', ('Y',)),)),
    ))
    print('EXPEC:', expected)
    assert third == expected

    print('FOURTH:', fourth)
    expected = ('rule', 'att', ('Y',), (
        ('term', 'rel', ('_', 'Y')),
        ('forall', 'rel', ('X', 'Y'), (('term', 'obj', ('X',)),)),
    ))
    print('EXPECT:', expected)
    assert fourth == expected
